"""
Global setting of the trading platform.
"""

from logging import INFO
from tzlocal import get_localzone_name

from .utility import load_json


SETTINGS: dict[str, object] = {
    "font.family": "",
    "font.size": 12,

    "log.active": True,
    "log.level": INFO,
    "log.console": True,
    "log.file": True,

    "email.server": "",
    "email.port": 0,
    "email.username": "",
    "email.password": "",
    "email.sender": "",
    "email.receiver": "",

    "datafeed.name": "",
    "datafeed.username": "",
    "datafeed.password": "",

    "database.timezone": get_localzone_name(),
    "database.name": "duckdb",
    "database.database": "database.duckdb",
    "database.host": "",
    "database.port": 0,
    "database.user": "",
    "database.password": "",

    "telegram.active": False,
    "telegram.token": "",
    "telegram.chat": 0,
    "telegram.proxy": "",
}


# Load global setting from json file.
SETTING_FILENAME: str = "vt_setting.json"
SETTINGS.update(load_json(SETTING_FILENAME))


def get_settings(prefix: str = "") -> dict[str, object]:
    prefix_length: int = len(prefix)
    settings = {k[prefix_length:]: v for k, v in SETTINGS.items() if k.startswith(prefix)}
    return settings



import sys
import traceback
from datetime import datetime
from enum import Enum
from multiprocessing.dummy import Pool
from queue import Empty, Queue
from typing import Callable, Optional, Union, Type
from types import TracebackType

import requests


CALLBACK_TYPE = Callable[[dict, "Request"], object]
ON_FAILED_TYPE = Callable[[int, "Request"], object]
ON_ERROR_TYPE = Callable[[Type, Exception, TracebackType, "Request"], object]


class RequestStatus(Enum):
    """
    Request status enum.
    """

    ready = 0       # Request created
    success = 1     # Request successful (status code 2xx)
    failed = 2      # Request failed (status code not 2xx)
    error = 3       # Exception raised


class Request(object):
    """
    Request object for status check.
    """

    def __init__(
        self,
        method: str,
        path: str,
        params: dict,
        data: Union[dict, str, bytes],
        headers: dict,
        callback: CALLBACK_TYPE = None,
        on_failed: ON_FAILED_TYPE = None,
        on_error: ON_ERROR_TYPE = None,
        extra: object = None,
    ) -> None:
        """"""
        self.method: str = method
        self.path: str = path
        self.callback: CALLBACK_TYPE = callback
        self.params: dict = params
        self.data: Union[dict, str, bytes] = data
        self.headers: dict = headers

        self.on_failed: ON_FAILED_TYPE = on_failed
        self.on_error: ON_ERROR_TYPE = on_error
        self.extra: object = extra

        self.response: requests.Response = None
        self.status: RequestStatus = RequestStatus.ready

    def __str__(self) -> str:
        """"""
        if self.response is None:
            status_code = "terminated"
        else:
            status_code = self.response.status_code

        return (
            "request : {} {} {} because {}: \n"
            "headers: {}\n"
            "params: {}\n"
            "data: {}\n"
            "response:"
            "{}\n".format(
                self.method,
                self.path,
                self.status.name,
                status_code,
                self.headers,
                self.params,
                self.data,
                "" if self.response is None else self.response.text,
            )
        )


class RestClient(object):
    """
    HTTP Client designed for all sorts of trading RESTFul API.
    * Reimplement sign function to add signature function.
    * Reimplement on_failed function to handle Non-2xx responses.
    * Use on_failed parameter in add_request function for individual Non-2xx response handling.
    * Reimplement on_error function to handle exception msg.
    """

    def __init__(self) -> None:
        """"""
        self.url_base: str = ""
        self.active: bool = False

        self.queue: Queue = Queue()
        self.pool: Pool = None

        self.proxies: dict = None

    def init(
        self,
        url_base: str,
        proxy_host: str = "",
        proxy_port: int = 0
    ) -> None:
        """
        Init rest client with url_base which is the API root address.
        """
        self.url_base = url_base

        if proxy_host and proxy_port:
            proxy: str = f"http://{proxy_host}:{proxy_port}"
            self.proxies = {"http": proxy, "https": proxy}

    def start(self, session_count: int = 5) -> None:
        """
        Start rest client with session_count number of threads.
        """
        if self.active:
            return

        self.active = True
        self.pool = Pool(session_count)
        self.pool.apply_async(self.run)

    def stop(self) -> None:
        """
        Stop rest client immediately.
        """
        self.active = False

    def join(self) -> None:
        """
        Wait till all requests are processed.
        """
        self.queue.join()

    def add_request(
        self,
        method: str,
        path: str,
        callback: CALLBACK_TYPE,
        params: dict = None,
        data: Union[dict, str, bytes] = None,
        headers: dict = None,
        on_failed: ON_FAILED_TYPE = None,
        on_error: ON_ERROR_TYPE = None,
        extra: object = None,
    ) -> Request:
        """
        Add a new request.
        :param method: GET, POST, PUT, DELETE, QUERY
        :param path: url path for query
        :param callback: callback function if 2xx status, type: (dict, Request)
        :param params: dict for query string
        :param data: Http body. If it is a dict, it will be converted to form-data. Otherwise, it will be converted to bytes.
        :param headers: dict for headers
        :param on_failed: callback function if Non-2xx status, type, type: (code, dict, Request)
        :param on_error: callback function when catching Python exception, type: (etype, evalue, tb, Request)
        :param extra: object extra data which can be used when handling callback
        :return: Request
        """
        request: Request = Request(
            method,
            path,
            params,
            data,
            headers,
            callback,
            on_failed,
            on_error,
            extra,
        )
        self.queue.put(request)
        return request

    def run(self) -> None:
        """"""
        try:
            session: requests.Session = requests.session()
            while self.active:
                try:
                    request: Request = self.queue.get(timeout=1)
                    try:
                        self.process_request(request, session)
                    finally:
                        self.queue.task_done()
                except Empty:
                    pass
        except Exception:
            et, ev, tb = sys.exc_info()
            self.on_error(et, ev, tb, None)

    def sign(self, request: Request) -> RequestStatus:
        """
        This function is called before sending object request out.
        Please implement signature method here.
        """
        return request

    def on_failed(self, status_code: int, request: Request) -> None:
        """
        Default on_failed handler for Non-2xx response.
        """
        print("RestClient on failed" + "-" * 10)
        print(str(request))

    def on_error(
        self,
        exception_type: type,
        exception_value: Exception,
        tb: TracebackType,
        request: Optional[Request],
    ) -> None:
        """
        Default on_error handler for Python exception.
        """
        try:
            print("RestClient on error" + "-" * 10)
            print(self.exception_detail(exception_type, exception_value, tb, request))
        except Exception:
            traceback.print_exc()

    def exception_detail(
        self,
        exception_type: type,
        exception_value: Exception,
        tb: TracebackType,
        request: Optional[Request],
    ) -> str:
        text: str = "[{}]: Unhandled RestClient Error:{}\n".format(
            datetime.now().isoformat(), exception_type
        )
        text += "request:{}\n".format(request)
        text += "Exception trace: \n"
        text += "".join(
            traceback.format_exception(exception_type, exception_value, tb)
        )
        return text

    def process_request(self, request: Request, session: requests.Session) -> None:
        """
        Sending request to server and get result.
        """
        try:
            request = self.sign(request)

            url: str = self.make_full_url(request.path)

            response: requests.Response = session.request(
                request.method,
                url,
                headers=request.headers,
                params=request.params,
                data=request.data,
                proxies=self.proxies,
            )
            request.response = response

            status_code: int = response.status_code
            if status_code // 100 == 2:  # 2xx codes are all successful
                json_body: dict = None
                if status_code != 204:
                    json_body = response.json()

                request.callback(json_body, request)
                request.status = RequestStatus.success
            else:
                request.status = RequestStatus.failed

                if request.on_failed:
                    request.on_failed(status_code, request)
                else:
                    self.on_failed(status_code, request)
        except Exception:
            request.status = RequestStatus.error
            t, v, tb = sys.exc_info()
            if request.on_error:
                request.on_error(t, v, tb, request)
            else:
                self.on_error(t, v, tb, request)

    def make_full_url(self, path: str) -> str:
        """
        Make relative api path into full url.
        eg: make_full_url("/get") == "http://xxxxx/get"
        """
        url = self.url_base + path
        return url

    def request(
        self,
        method: str,
        path: str,
        params: dict = None,
        data: dict = None,
        headers: dict = None,
    ) -> requests.Response:
        """
        Add a new request.
        :param method: GET, POST, PUT, DELETE, QUERY
        :param path: url path for query
        :param params: dict for query string
        :param data: dict for body
        :param headers: dict for headers
        :return: requests.Response
        """
        request: Request = Request(
            method,
            path,
            params,
            data,
            headers
        )
        request = self.sign(request)

        url: str = self.make_full_url(request.path)

        response: requests.Response = requests.request(
            request.method,
            url,
            headers=request.headers,
            params=request.params,
            data=request.data,
            proxies=self.proxies,
        )
        return response



"""
Basic widgets for UI.
"""

import csv
from datetime import datetime
from enum import Enum
from tzlocal import get_localzone_name

from qfluentwidgets import TableWidget, PushButton, LineEdit

from vnpy.trader.locale import _
from .qt import QtCore, QtGui, QtWidgets
from ..constant import Direction
from ..engine import MainEngine, Event, EventEngine
from ..event import (
    EVENT_QUOTE,
    EVENT_TICK,
    EVENT_TRADE,
    EVENT_ORDER,
    EVENT_POSITION,
    EVENT_ACCOUNT,
    EVENT_LOG
)
from ..object import (
    CancelRequest,
    ContractData,
    OrderData,
    QuoteData,
)
from ..utility import ZoneInfo


COLOR_LONG = QtGui.QColor("red")
COLOR_SHORT = QtGui.QColor("green")
COLOR_BID = QtGui.QColor("red")
COLOR_ASK = QtGui.QColor("green")
COLOR_BLACK = QtGui.QColor("black")


class BaseCell(QtWidgets.QTableWidgetItem):
    """
    General cell used in tablewidgets.
    """

    def __init__(self, content: object, data: object) -> None:
        """"""
        super().__init__()
        self.setTextAlignment(QtCore.Qt.AlignCenter)
        self.set_content(content, data)

    def set_content(self, content: object, data: object) -> None:
        """
        Set text content.
        """
        self.setText(str(content))
        self._data = data

    def get_data(self) -> object:
        """
        Get data object.
        """
        return self._data


class EnumCell(BaseCell):
    """
    Cell used for showing enum data.
    """

    def __init__(self, content: str, data: object) -> None:
        """"""
        super().__init__(content, data)

    def set_content(self, content: object, data: object) -> None:
        """
        Set text using enum.constant.value.
        """
        if content:
            super().set_content(content.value, data)


class DirectionCell(EnumCell):
    """
    Cell used for showing direction data.
    """

    def __init__(self, content: str, data: object) -> None:
        """"""
        super().__init__(content, data)

    def set_content(self, content: object, data: object) -> None:
        """
        Cell color is set according to direction.
        """
        super().set_content(content, data)

        if content is Direction.SHORT:
            self.setForeground(COLOR_SHORT)
        else:
            self.setForeground(COLOR_LONG)


class BidCell(BaseCell):
    """
    Cell used for showing bid price and volume.
    """

    def __init__(self, content: object, data: object) -> None:
        """"""
        super().__init__(content, data)

        self.setForeground(COLOR_BID)


class AskCell(BaseCell):
    """
    Cell used for showing ask price and volume.
    """

    def __init__(self, content: object, data: object) -> None:
        """"""
        super().__init__(content, data)

        self.setForeground(COLOR_ASK)


class PnlCell(BaseCell):
    """
    Cell used for showing pnl data.
    """

    def __init__(self, content: object, data: object) -> None:
        """"""
        super().__init__(content, data)

    def set_content(self, content: object, data: object) -> None:
        """
        Cell color is set based on whether pnl is
        positive or negative.
        """
        super().set_content(content, data)

        if str(content).startswith("-"):
            self.setForeground(COLOR_SHORT)
        else:
            self.setForeground(COLOR_LONG)


class TimeCell(BaseCell):
    """
    Cell used for showing time string from datetime object.
    """

    local_tz = ZoneInfo(get_localzone_name())

    def __init__(self, content: object, data: object) -> None:
        """"""
        super().__init__(content, data)

    def set_content(self, content: object, data: object) -> None:
        """"""
        if content is None:
            return

        content: datetime = content.astimezone(self.local_tz)
        timestamp: str = content.strftime("%H:%M:%S")

        millisecond: int = int(content.microsecond / 1000)
        if millisecond:
            timestamp = f"{timestamp}.{millisecond}"
        else:
            timestamp = f"{timestamp}.000"

        self.setText(timestamp)
        self._data = data


class DateCell(BaseCell):
    """
    Cell used for showing date string from datetime object.
    """

    def __init__(self, content: object, data: object) -> None:
        """"""
        super().__init__(content, data)

    def set_content(self, content: object, data: object) -> None:
        """"""
        if content is None:
            return

        self.setText(content.strftime("%Y-%m-%d"))
        self._data = data


class MsgCell(BaseCell):
    """
    Cell used for showing msg data.
    """

    def __init__(self, content: str, data: object) -> None:
        """"""
        super().__init__(content, data)
        self.setTextAlignment(QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter)


class BaseMonitor(TableWidget):
    """
    Monitor data update.
    """

    event_type: str = ""
    data_key: str = ""
    sorting: bool = False
    headers: dict = {}

    signal: QtCore.Signal = QtCore.Signal(Event)

    def __init__(self, main_engine: MainEngine, event_engine: EventEngine) -> None:
        """"""
        super().__init__()

        self.main_engine: MainEngine = main_engine
        self.event_engine: EventEngine = event_engine
        self.cells: dict[str, dict] = {}

        self.init_ui()
        self.load_setting()
        self.register_event()

    def init_ui(self) -> None:
        """"""
        self.init_table()
        self.init_menu()

    def init_table(self) -> None:
        """
        Initialize table.
        """
        self.setColumnCount(len(self.headers))

        labels: list = [d["display"] for d in self.headers.values()]
        self.setHorizontalHeaderLabels(labels)

        self.verticalHeader().setVisible(False)
        self.setEditTriggers(self.EditTrigger.NoEditTriggers)
        self.setAlternatingRowColors(True)
        self.setSortingEnabled(self.sorting)

    def init_menu(self) -> None:
        """
        Create right click menu.
        """
        self.menu: QtWidgets.QMenu = QtWidgets.QMenu(self)

        resize_action: QtGui.QAction = QtWidgets.QAction(_("调整列宽"), self)
        resize_action.triggered.connect(self.resize_columns)
        self.menu.addAction(resize_action)

        save_action: QtGui.QAction = QtWidgets.QAction(_("保存数据"), self)
        save_action.triggered.connect(self.save_csv)
        self.menu.addAction(save_action)

    def register_event(self) -> None:
        """
        Register event handler into event engine.
        """
        if self.event_type:
            self.signal.connect(self.process_event)
            self.event_engine.register(self.event_type, self.signal.emit)

    def process_event(self, event: Event) -> None:
        """
        Process new data from event and update into table.
        """
        # Disable sorting to prevent unwanted error.
        if self.sorting:
            self.setSortingEnabled(False)

        # Update data into table.
        data = event.data

        if not self.data_key:
            self.insert_new_row(data)
        else:
            key: str = data.__getattribute__(self.data_key)

            if key in self.cells:
                self.update_old_row(data)
            else:
                self.insert_new_row(data)

        # Enable sorting
        if self.sorting:
            self.setSortingEnabled(True)

    def insert_new_row(self, data: object) -> None:
        """
        Insert a new row at the top of table.
        """
        self.insertRow(0)

        row_cells: dict = {}
        for column, header in enumerate(self.headers.keys()):
            setting: dict = self.headers[header]

            content = data.__getattribute__(header)
            cell: QtWidgets.QTableWidgetItem = setting["cell"](content, data)
            self.setItem(0, column, cell)

            if setting["update"]:
                row_cells[header] = cell

        if self.data_key:
            key: str = data.__getattribute__(self.data_key)
            self.cells[key] = row_cells

    def update_old_row(self, data: object) -> None:
        """
        Update an old row in table.
        """
        key: str = data.__getattribute__(self.data_key)
        row_cells = self.cells[key]

        for header, cell in row_cells.items():
            content = data.__getattribute__(header)
            cell.set_content(content, data)

    def resize_columns(self) -> None:
        """
        Resize all columns according to contents.
        """
        self.horizontalHeader().resizeSections(QtWidgets.QHeaderView.ResizeToContents)

    def save_csv(self) -> None:
        """
        Save table data into a csv file
        """
        path, __ = QtWidgets.QFileDialog.getSaveFileName(
            self, _("保存数据"), "", "CSV(*.csv)")

        if not path:
            return

        with open(path, "w") as f:
            writer = csv.writer(f, lineterminator="\n")

            headers: list = [d["display"] for d in self.headers.values()]
            writer.writerow(headers)

            for row in range(self.rowCount()):
                if self.isRowHidden(row):
                    continue

                row_data: list = []
                for column in range(self.columnCount()):
                    item: QtWidgets.QTableWidgetItem = self.item(row, column)
                    if item:
                        row_data.append(str(item.text()))
                    else:
                        row_data.append("")
                writer.writerow(row_data)

    def contextMenuEvent(self, event: QtGui.QContextMenuEvent) -> None:
        """
        Show menu with right click.
        """
        self.menu.popup(QtGui.QCursor.pos())

    def save_setting(self) -> None:
        """"""
        settings: QtCore.QSettings = QtCore.QSettings(self.__class__.__name__, "custom")
        settings.setValue("column_state", self.horizontalHeader().saveState())

    def load_setting(self) -> None:
        """"""
        settings: QtCore.QSettings = QtCore.QSettings(self.__class__.__name__, "custom")
        column_state = settings.value("column_state")

        if isinstance(column_state, QtCore.QByteArray):
            self.horizontalHeader().restoreState(column_state)
            self.horizontalHeader().setSortIndicator(-1, QtCore.Qt.AscendingOrder)


class TickMonitor(BaseMonitor):
    """
    Monitor for tick data.
    """

    event_type: str = EVENT_TICK
    data_key: str = "vt_symbol"
    sorting: bool = True

    headers: dict = {
        "symbol": {"display": _("代码"), "cell": BaseCell, "update": False},
        "exchange": {"display": _("交易所"), "cell": EnumCell, "update": False},
        "name": {"display": _("名称"), "cell": BaseCell, "update": True},
        "last_price": {"display": _("最新价"), "cell": BaseCell, "update": True},
        "volume": {"display": _("成交量"), "cell": BaseCell, "update": True},
        "open_price": {"display": _("开盘价"), "cell": BaseCell, "update": True},
        "high_price": {"display": _("最高价"), "cell": BaseCell, "update": True},
        "low_price": {"display": _("最低价"), "cell": BaseCell, "update": True},
        "bid_price_1": {"display": _("买1价"), "cell": BidCell, "update": True},
        "bid_volume_1": {"display": _("买1量"), "cell": BidCell, "update": True},
        "ask_price_1": {"display": _("卖1价"), "cell": AskCell, "update": True},
        "ask_volume_1": {"display": _("卖1量"), "cell": AskCell, "update": True},
        "datetime": {"display": _("时间"), "cell": TimeCell, "update": True},
        "gateway_name": {"display": _("接口"), "cell": BaseCell, "update": False},
    }


class LogMonitor(BaseMonitor):
    """
    Monitor for log data.
    """

    event_type: str = EVENT_LOG
    data_key: str = ""
    sorting: bool = False

    headers: dict = {
        "time": {"display": _("时间"), "cell": TimeCell, "update": False},
        "msg": {"display": _("信息"), "cell": MsgCell, "update": False},
        "gateway_name": {"display": _("接口"), "cell": BaseCell, "update": False},
    }


class TradeMonitor(BaseMonitor):
    """
    Monitor for trade data.
    """

    event_type: str = EVENT_TRADE
    data_key: str = ""
    sorting: bool = True

    headers: dict = {
        "tradeid": {"display": _("成交号"), "cell": BaseCell, "update": False},
        "orderid": {"display": _("委托号"), "cell": BaseCell, "update": False},
        "symbol": {"display": _("代码"), "cell": BaseCell, "update": False},
        "exchange": {"display": _("交易所"), "cell": EnumCell, "update": False},
        "direction": {"display": _("方向"), "cell": DirectionCell, "update": False},
        "offset": {"display": _("开平"), "cell": EnumCell, "update": False},
        "price": {"display": _("价格"), "cell": BaseCell, "update": False},
        "volume": {"display": _("数量"), "cell": BaseCell, "update": False},
        "datetime": {"display": _("时间"), "cell": TimeCell, "update": False},
        "gateway_name": {"display": _("接口"), "cell": BaseCell, "update": False},
    }


class OrderMonitor(BaseMonitor):
    """
    Monitor for order data.
    """

    event_type: str = EVENT_ORDER
    data_key: str = "vt_orderid"
    sorting: bool = True

    headers: dict = {
        "orderid": {"display": _("委托号"), "cell": BaseCell, "update": False},
        "reference": {"display": _("来源"), "cell": BaseCell, "update": False},
        "symbol": {"display": _("代码"), "cell": BaseCell, "update": False},
        "exchange": {"display": _("交易所"), "cell": EnumCell, "update": False},
        "type": {"display": _("类型"), "cell": EnumCell, "update": False},
        "direction": {"display": _("方向"), "cell": DirectionCell, "update": False},
        "offset": {"display": _("开平"), "cell": EnumCell, "update": False},
        "price": {"display": _("价格"), "cell": BaseCell, "update": False},
        "volume": {"display": _("总数量"), "cell": BaseCell, "update": True},
        "traded": {"display": _("已成交"), "cell": BaseCell, "update": True},
        "status": {"display": _("状态"), "cell": EnumCell, "update": True},
        "datetime": {"display": _("时间"), "cell": TimeCell, "update": True},
        "gateway_name": {"display": _("接口"), "cell": BaseCell, "update": False},
    }

    def init_ui(self) -> None:
        """
        Connect signal.
        """
        super().init_ui()

        self.setToolTip(_("双击单元格撤单"))
        self.itemDoubleClicked.connect(self.cancel_order)

    def cancel_order(self, cell: BaseCell) -> None:
        """
        Cancel order if cell double clicked.
        """
        order: OrderData = cell.get_data()
        req: CancelRequest = order.create_cancel_request()
        self.main_engine.cancel_order(req, order.gateway_name)


class PositionMonitor(BaseMonitor):
    """
    Monitor for position data.
    """

    event_type: str = EVENT_POSITION
    data_key: str = "vt_positionid"
    sorting: bool = True

    headers: dict = {
        "symbol": {"display": _("代码"), "cell": BaseCell, "update": False},
        "exchange": {"display": _("交易所"), "cell": EnumCell, "update": False},
        "direction": {"display": _("方向"), "cell": DirectionCell, "update": False},
        "volume": {"display": _("数量"), "cell": BaseCell, "update": True},
        "yd_volume": {"display": _("昨仓"), "cell": BaseCell, "update": True},
        "frozen": {"display": _("冻结"), "cell": BaseCell, "update": True},
        "price": {"display": _("均价"), "cell": BaseCell, "update": True},
        "pnl": {"display": _("盈亏"), "cell": PnlCell, "update": True},
        "gateway_name": {"display": _("接口"), "cell": BaseCell, "update": False},
    }


class AccountMonitor(BaseMonitor):
    """
    Monitor for account data.
    """

    event_type: str = EVENT_ACCOUNT
    data_key: str = "vt_accountid"
    sorting: bool = True

    headers: dict = {
        "accountid": {"display": _("账号"), "cell": BaseCell, "update": False},
        "balance": {"display": _("余额"), "cell": BaseCell, "update": True},
        "frozen": {"display": _("冻结"), "cell": BaseCell, "update": True},
        "available": {"display": _("可用"), "cell": BaseCell, "update": True},
        "gateway_name": {"display": _("接口"), "cell": BaseCell, "update": False},
    }


class QuoteMonitor(BaseMonitor):
    """
    Monitor for quote data.
    """

    event_type: str = EVENT_QUOTE
    data_key: str = "vt_quoteid"
    sorting: bool = True

    headers: dict = {
        "quoteid": {"display": _("报价号"), "cell": BaseCell, "update": False},
        "reference": {"display": _("来源"), "cell": BaseCell, "update": False},
        "symbol": {"display": _("代码"), "cell": BaseCell, "update": False},
        "exchange": {"display": _("交易所"), "cell": EnumCell, "update": False},
        "bid_offset": {"display": _("买开平"), "cell": EnumCell, "update": False},
        "bid_volume": {"display": _("买量"), "cell": BidCell, "update": False},
        "bid_price": {"display": _("买价"), "cell": BidCell, "update": False},
        "ask_price": {"display": _("卖价"), "cell": AskCell, "update": False},
        "ask_volume": {"display": _("卖量"), "cell": AskCell, "update": False},
        "ask_offset": {"display": _("卖开平"), "cell": EnumCell, "update": False},
        "status": {"display": _("状态"), "cell": EnumCell, "update": True},
        "datetime": {"display": _("时间"), "cell": TimeCell, "update": True},
        "gateway_name": {"display": _("接口"), "cell": BaseCell, "update": False},
    }

    def init_ui(self):
        """
        Connect signal.
        """
        super().init_ui()

        self.setToolTip(_("双击单元格撤销报价"))
        self.itemDoubleClicked.connect(self.cancel_quote)

    def cancel_quote(self, cell: BaseCell) -> None:
        """
        Cancel quote if cell double clicked.
        """
        quote: QuoteData = cell.get_data()
        req: CancelRequest = quote.create_cancel_request()
        self.main_engine.cancel_quote(req, quote.gateway_name)


class ActiveOrderMonitor(OrderMonitor):
    """
    Monitor which shows active order only.
    """

    def process_event(self, event) -> None:
        """
        Hides the row if order is not active.
        """
        super().process_event(event)

        order: OrderData = event.data
        row_cells: dict = self.cells[order.vt_orderid]
        row: int = self.row(row_cells["volume"])

        if order.is_active():
            self.showRow(row)
        else:
            self.hideRow(row)


class ContractManager(QtWidgets.QWidget):
    """
    Query contract data available to trade in system.
    """

    headers: dict[str, str] = {
        "vt_symbol": _("本地代码"),
        "symbol": _("代码"),
        "exchange": _("交易所"),
        "name": _("名称"),
        "product": _("合约分类"),
        "size": _("合约乘数"),
        "pricetick": _("价格跳动"),
        "min_volume": _("最小委托量"),
        "option_portfolio": _("期权产品"),
        "option_expiry": _("期权到期日"),
        "option_strike": _("期权行权价"),
        "option_type": _("期权类型"),
        "gateway_name": _("交易接口"),
    }

    def __init__(self, main_engine: MainEngine, event_engine: EventEngine) -> None:
        super().__init__()

        self.main_engine: MainEngine = main_engine
        self.event_engine: EventEngine = event_engine

        self.init_ui()

    def init_ui(self) -> None:
        """"""
        self.setWindowTitle(_("合约查询"))
        self.resize(1000, 600)

        self.filter_line: LineEdit = LineEdit()
        self.filter_line.setPlaceholderText(_("输入合约代码或者交易所，留空则查询所有合约"))

        self.button_show: PushButton = PushButton(_("查询"))
        self.button_show.clicked.connect(self.show_contracts)

        labels: list = []
        for name, display in self.headers.items():
            label: str = f"{display}"
            labels.append(label)

        self.contract_table: TableWidget = TableWidget()
        self.contract_table.setColumnCount(len(self.headers))
        self.contract_table.setHorizontalHeaderLabels(labels)
        self.contract_table.verticalHeader().setVisible(False)
        self.contract_table.setEditTriggers(self.contract_table.EditTrigger.NoEditTriggers)
        self.contract_table.setAlternatingRowColors(True)

        hbox: QtWidgets.QHBoxLayout = QtWidgets.QHBoxLayout()
        hbox.addWidget(self.filter_line)
        hbox.addWidget(self.button_show)

        vbox: QtWidgets.QVBoxLayout = QtWidgets.QVBoxLayout()
        vbox.addLayout(hbox)
        vbox.addWidget(self.contract_table)

        self.setLayout(vbox)

    def show_contracts(self) -> None:
        """
        Show contracts by symbol
        """
        flt: str = str(self.filter_line.text())

        all_contracts: list[ContractData] = self.main_engine.get_all_contracts()
        if flt:
            contracts: list[ContractData] = [
                contract for contract in all_contracts if flt in contract.vt_symbol
            ]
        else:
            contracts: list[ContractData] = all_contracts

        self.contract_table.clearContents()
        self.contract_table.setRowCount(len(contracts))

        for row, contract in enumerate(contracts):
            for column, name in enumerate(self.headers.keys()):
                value: object = getattr(contract, name)

                if value in {None, 0, 0.0}:
                    value = ""

                if isinstance(value, Enum):
                    cell: EnumCell = EnumCell(value, contract)
                elif isinstance(value, datetime):
                    cell: DateCell = DateCell(value, contract)
                else:
                    cell: BaseCell = BaseCell(value, contract)
                self.contract_table.setItem(row, column, cell)

        self.contract_table.resizeColumnsToContents()



"""
Basic widgets for UI.
"""

import platform
from copy import copy
from typing import TYPE_CHECKING

import importlib_metadata
from qfluentwidgets import (
    Pivot, MessageBoxBase, BodyLabel, SubtitleLabel,
    PushButton, ComboBox, LineEdit, CheckBox,
)

from vnpy.trader.locale import _
from .qt import QtCore, QtGui, QtWidgets
from ..constant import Direction, Exchange, Offset, OrderType
from ..engine import MainEngine, Event, EventEngine
from ..event import EVENT_TICK
from ..object import (
    OrderRequest,
    SubscribeRequest,
    CancelRequest,
    ContractData,
    PositionData,
    OrderData,
    TickData
)
from ..utility import load_json, save_json, get_digits
from ..setting import SETTING_FILENAME, SETTINGS

if TYPE_CHECKING:
    from .monitor import BaseCell


class ConnectDialog(MessageBoxBase):
    """
    Start connection of a certain gateway.
    """

    def __init__(self, main_engine: MainEngine, gateway_name: str, parent: QtWidgets.QWidget = None) -> None:
        """"""
        super().__init__(parent)

        self.main_engine: MainEngine = main_engine
        self.gateway_name: str = gateway_name
        self.filename: str = f"connect_{gateway_name.lower()}.json"

        self.widgets: dict[str, QtWidgets.QWidget] = {}

        self.init_ui()

    def init_ui(self) -> None:
        """"""
        self.title_label = SubtitleLabel(f"Connect {self.gateway_name}", self)

        # Default setting provides field name, field data type and field default value.
        default_setting: dict = self.main_engine.get_default_setting(
            self.gateway_name)

        # Saved setting provides field data used last time.
        loaded_setting: dict = load_json(self.filename)

        # Initialize line edits and form layout based on setting.
        grid: QtWidgets.QGridLayout = QtWidgets.QGridLayout()
        row: int = 0

        for field_name, field_value in default_setting.items():
            field_type: type = type(field_value)

            if field_type == list:
                widget: ComboBox = ComboBox()
                widget.addItems(field_value)

                if field_name in loaded_setting:
                    saved_value = loaded_setting[field_name]
                    ix: int = widget.findText(saved_value)
                    widget.setCurrentIndex(ix)
            else:
                widget: LineEdit = LineEdit()
                widget.setText(str(field_value))

                if field_name in loaded_setting:
                    saved_value = loaded_setting[field_name]
                    widget.setText(str(saved_value))

                if _("密码") in field_name:
                    widget.setEchoMode(LineEdit.Password)

                if field_type == int:
                    validator: QtGui.QIntValidator = QtGui.QIntValidator()
                    widget.setValidator(validator)

            grid.addWidget(BodyLabel(f"{field_name} <{field_type.__name__}>"), row, 0)
            grid.addWidget(widget, row, 1)
            self.widgets[field_name] = (widget, field_type)

            row += 1

        self.viewLayout.addWidget(self.title_label)
        self.viewLayout.addLayout(grid)

        self.yesButton.setText("Connect")
        self.yesButton.clicked.connect(self.connect_gateway)

        self.cancelButton.setText("Cancel")

        self.widget.setFixedWidth(self.widget.width() * 6)

    def connect_gateway(self) -> None:
        """
        Get setting value from line edits and connect the gateway.
        """
        setting: dict = {}
        for field_name, tp in self.widgets.items():
            widget, field_type = tp
            if field_type == list:
                field_value = str(widget.currentText())
            else:
                try:
                    field_value = field_type(widget.text())
                except ValueError:
                    field_value = field_type()
            setting[field_name] = field_value

        save_json(self.filename, setting)

        self.main_engine.connect(setting, self.gateway_name)
        self.accept()


class TradingWidget(QtWidgets.QWidget):
    """
    General manual trading widget.
    """

    signal_tick: QtCore.Signal = QtCore.Signal(Event)

    def __init__(self, main_engine: MainEngine, event_engine: EventEngine) -> None:
        """"""
        super().__init__()

        self.main_engine: MainEngine = main_engine
        self.event_engine: EventEngine = event_engine

        self.vt_symbol: str = ""
        self.price_digits: int = 0

        self.init_ui()
        self.register_event()

    def init_ui(self) -> None:
        """"""
        self.setFixedWidth(300)

        # Trading function area
        exchanges: list[Exchange] = self.main_engine.get_all_exchanges()
        self.exchange_combo: ComboBox = ComboBox()
        self.exchange_combo.addItems([exchange.value for exchange in exchanges])

        self.symbol_line: LineEdit = LineEdit()
        self.symbol_line.returnPressed.connect(self.set_vt_symbol)

        self.name_line: LineEdit = LineEdit()
        self.name_line.setReadOnly(True)

        self.direction_combo: ComboBox = ComboBox()
        self.direction_combo.addItems([Direction.LONG.value, Direction.SHORT.value])

        self.offset_combo: ComboBox = ComboBox()
        self.offset_combo.addItems([offset.value for offset in Offset])

        self.order_type_combo: ComboBox = ComboBox()
        self.order_type_combo.addItems([order_type.value for order_type in OrderType])

        double_validator: QtGui.QDoubleValidator = QtGui.QDoubleValidator()
        double_validator.setBottom(0)

        self.price_line: LineEdit = LineEdit()
        self.price_line.setValidator(double_validator)

        self.volume_line: LineEdit = LineEdit()
        self.volume_line.setValidator(double_validator)

        self.gateway_combo: ComboBox = ComboBox()
        self.gateway_combo.addItems(self.main_engine.get_all_gateway_names())

        self.price_check: CheckBox = CheckBox()
        self.price_check.setToolTip(_("设置价格随行情更新"))

        send_button: PushButton = PushButton("Send Order")
        send_button.clicked.connect(self.send_order)

        cancel_button: PushButton = PushButton("Cancel All")
        cancel_button.clicked.connect(self.cancel_all)

        grid: QtWidgets.QGridLayout = QtWidgets.QGridLayout()
        grid.addWidget(BodyLabel(_("交易所")), 0, 0)
        grid.addWidget(BodyLabel(_("代码")), 1, 0)
        grid.addWidget(BodyLabel(_("名称")), 2, 0)
        grid.addWidget(BodyLabel(_("方向")), 3, 0)
        grid.addWidget(BodyLabel(_("开平")), 4, 0)
        grid.addWidget(BodyLabel(_("类型")), 5, 0)
        grid.addWidget(BodyLabel(_("价格")), 6, 0)
        grid.addWidget(BodyLabel(_("数量")), 7, 0)
        grid.addWidget(BodyLabel(_("接口")), 8, 0)
        grid.addWidget(self.exchange_combo, 0, 1, 1, 2)
        grid.addWidget(self.symbol_line, 1, 1, 1, 2)
        grid.addWidget(self.name_line, 2, 1, 1, 2)
        grid.addWidget(self.direction_combo, 3, 1, 1, 2)
        grid.addWidget(self.offset_combo, 4, 1, 1, 2)
        grid.addWidget(self.order_type_combo, 5, 1, 1, 2)
        grid.addWidget(self.price_line, 6, 1, 1, 1)
        grid.addWidget(self.price_check, 6, 2, 1, 1)
        grid.addWidget(self.volume_line, 7, 1, 1, 2)
        grid.addWidget(self.gateway_combo, 8, 1, 1, 2)
        grid.addWidget(send_button, 9, 0, 1, 3)
        grid.addWidget(cancel_button, 10, 0, 1, 3)

        # Market depth display area
        bid_color: str = "red"
        ask_color: str = "green"

        self.bp1_label: BodyLabel = self.create_label(bid_color)
        self.bp2_label: BodyLabel = self.create_label(bid_color)
        self.bp3_label: BodyLabel = self.create_label(bid_color)
        self.bp4_label: BodyLabel = self.create_label(bid_color)
        self.bp5_label: BodyLabel = self.create_label(bid_color)

        self.bv1_label: BodyLabel = self.create_label(
            bid_color, alignment=QtCore.Qt.AlignRight)
        self.bv2_label: BodyLabel = self.create_label(
            bid_color, alignment=QtCore.Qt.AlignRight)
        self.bv3_label: BodyLabel = self.create_label(
            bid_color, alignment=QtCore.Qt.AlignRight)
        self.bv4_label: BodyLabel = self.create_label(
            bid_color, alignment=QtCore.Qt.AlignRight)
        self.bv5_label: BodyLabel = self.create_label(
            bid_color, alignment=QtCore.Qt.AlignRight)

        self.ap1_label: BodyLabel = self.create_label(ask_color)
        self.ap2_label: BodyLabel = self.create_label(ask_color)
        self.ap3_label: BodyLabel = self.create_label(ask_color)
        self.ap4_label: BodyLabel = self.create_label(ask_color)
        self.ap5_label: BodyLabel = self.create_label(ask_color)

        self.av1_label: BodyLabel = self.create_label(
            ask_color, alignment=QtCore.Qt.AlignRight)
        self.av2_label: BodyLabel = self.create_label(
            ask_color, alignment=QtCore.Qt.AlignRight)
        self.av3_label: BodyLabel = self.create_label(
            ask_color, alignment=QtCore.Qt.AlignRight)
        self.av4_label: BodyLabel = self.create_label(
            ask_color, alignment=QtCore.Qt.AlignRight)
        self.av5_label: BodyLabel = self.create_label(
            ask_color, alignment=QtCore.Qt.AlignRight)

        self.lp_label: BodyLabel = self.create_label()
        self.return_label: BodyLabel = self.create_label(alignment=QtCore.Qt.AlignRight)

        form: QtWidgets.QFormLayout = QtWidgets.QFormLayout()
        form.addRow(self.ap5_label, self.av5_label)
        form.addRow(self.ap4_label, self.av4_label)
        form.addRow(self.ap3_label, self.av3_label)
        form.addRow(self.ap2_label, self.av2_label)
        form.addRow(self.ap1_label, self.av1_label)
        form.addRow(self.lp_label, self.return_label)
        form.addRow(self.bp1_label, self.bv1_label)
        form.addRow(self.bp2_label, self.bv2_label)
        form.addRow(self.bp3_label, self.bv3_label)
        form.addRow(self.bp4_label, self.bv4_label)
        form.addRow(self.bp5_label, self.bv5_label)

        # Overall layout
        vbox: QtWidgets.QVBoxLayout = QtWidgets.QVBoxLayout()
        vbox.addLayout(grid)
        vbox.addLayout(form)
        self.setLayout(vbox)

    def create_label(
        self,
        color: str = "",
        alignment: int = QtCore.Qt.AlignLeft
    ) -> BodyLabel:
        """
        Create label with certain font color.
        """
        label: BodyLabel = BodyLabel()
        if color:
            label.setStyleSheet(f"color:{color}")
        label.setAlignment(alignment)
        return label

    def register_event(self) -> None:
        """"""
        self.signal_tick.connect(self.process_tick_event)
        self.event_engine.register(EVENT_TICK, self.signal_tick.emit)

    def process_tick_event(self, event: Event) -> None:
        """"""
        tick: TickData = event.data
        if tick.vt_symbol != self.vt_symbol:
            return

        price_digits: int = self.price_digits

        self.lp_label.setText(f"{tick.last_price:.{price_digits}f}")
        self.bp1_label.setText(f"{tick.bid_price_1:.{price_digits}f}")
        self.bv1_label.setText(str(tick.bid_volume_1))
        self.ap1_label.setText(f"{tick.ask_price_1:.{price_digits}f}")
        self.av1_label.setText(str(tick.ask_volume_1))

        if tick.pre_close:
            r: float = (tick.last_price / tick.pre_close - 1) * 100
            self.return_label.setText(f"{r:.2f}%")

        if tick.bid_price_2:
            self.bp2_label.setText(f"{tick.bid_price_2:.{price_digits}f}")
            self.bv2_label.setText(str(tick.bid_volume_2))
            self.ap2_label.setText(f"{tick.ask_price_2:.{price_digits}f}")
            self.av2_label.setText(str(tick.ask_volume_2))

            self.bp3_label.setText(f"{tick.bid_price_3:.{price_digits}f}")
            self.bv3_label.setText(str(tick.bid_volume_3))
            self.ap3_label.setText(f"{tick.ask_price_3:.{price_digits}f}")
            self.av3_label.setText(str(tick.ask_volume_3))

            self.bp4_label.setText(f"{tick.bid_price_4:.{price_digits}f}")
            self.bv4_label.setText(str(tick.bid_volume_4))
            self.ap4_label.setText(f"{tick.ask_price_4:.{price_digits}f}")
            self.av4_label.setText(str(tick.ask_volume_4))

            self.bp5_label.setText(f"{tick.bid_price_5:.{price_digits}f}")
            self.bv5_label.setText(str(tick.bid_volume_5))
            self.ap5_label.setText(f"{tick.ask_price_5:.{price_digits}f}")
            self.av5_label.setText(str(tick.ask_volume_5))

        if self.price_check.isChecked():
            self.price_line.setText(f"{tick.last_price:.{price_digits}f}")

    def set_vt_symbol(self) -> None:
        """
        Set the tick depth data to monitor by vt_symbol.
        """
        symbol: str = str(self.symbol_line.text())
        if not symbol:
            return

        # Generate vt_symbol from symbol and exchange
        exchange_value: str = str(self.exchange_combo.currentText())
        vt_symbol: str = f"{symbol}.{exchange_value}"

        if vt_symbol == self.vt_symbol:
            return
        self.vt_symbol = vt_symbol

        # Update name line widget and clear all labels
        contract: ContractData = self.main_engine.get_contract(vt_symbol)
        if not contract:
            self.name_line.setText("")
            gateway_name: str = self.gateway_combo.currentText()
        else:
            self.name_line.setText(contract.name)
            gateway_name: str = contract.gateway_name

            # Update gateway combo box.
            ix: int = self.gateway_combo.findText(gateway_name)
            self.gateway_combo.setCurrentIndex(ix)

            # Update price digits
            self.price_digits = get_digits(contract.pricetick)

        self.clear_label_text()
        self.volume_line.setText("")
        self.price_line.setText("")

        # Subscribe tick data
        req: SubscribeRequest = SubscribeRequest(
            symbol=symbol, exchange=Exchange(exchange_value)
        )

        self.main_engine.subscribe(req, gateway_name)

    def clear_label_text(self) -> None:
        """
        Clear text on all labels.
        """
        self.lp_label.setText("")
        self.return_label.setText("")

        self.bv1_label.setText("")
        self.bv2_label.setText("")
        self.bv3_label.setText("")
        self.bv4_label.setText("")
        self.bv5_label.setText("")

        self.av1_label.setText("")
        self.av2_label.setText("")
        self.av3_label.setText("")
        self.av4_label.setText("")
        self.av5_label.setText("")

        self.bp1_label.setText("")
        self.bp2_label.setText("")
        self.bp3_label.setText("")
        self.bp4_label.setText("")
        self.bp5_label.setText("")

        self.ap1_label.setText("")
        self.ap2_label.setText("")
        self.ap3_label.setText("")
        self.ap4_label.setText("")
        self.ap5_label.setText("")

    def send_order(self) -> None:
        """
        Send new order manually.
        """
        symbol: str = str(self.symbol_line.text())
        if not symbol:
            QtWidgets.QMessageBox.critical(self, _("委托失败"), _("请输入合约代码"))
            return

        volume_text: str = str(self.volume_line.text())
        if not volume_text:
            QtWidgets.QMessageBox.critical(self, _("委托失败"), _("请输入委托数量"))
            return
        volume: float = float(volume_text)

        price_text: str = str(self.price_line.text())
        if not price_text:
            price = 0
        else:
            price = float(price_text)

        req: OrderRequest = OrderRequest(
            symbol=symbol,
            exchange=Exchange(str(self.exchange_combo.currentText())),
            direction=Direction(str(self.direction_combo.currentText())),
            type=OrderType(str(self.order_type_combo.currentText())),
            volume=volume,
            price=price,
            offset=Offset(str(self.offset_combo.currentText())),
            reference="ManualTrading"
        )

        gateway_name: str = str(self.gateway_combo.currentText())

        self.main_engine.send_order(req, gateway_name)

    def cancel_all(self) -> None:
        """
        Cancel all active orders.
        """
        order_list: list[OrderData] = self.main_engine.get_all_active_orders()
        for order in order_list:
            req: CancelRequest = order.create_cancel_request()
            self.main_engine.cancel_order(req, order.gateway_name)

    def update_with_cell(self, cell: "BaseCell") -> None:
        """"""
        data = cell.get_data()

        self.symbol_line.setText(data.symbol)
        self.exchange_combo.setCurrentIndex(
            self.exchange_combo.findText(data.exchange.value)
        )

        self.set_vt_symbol()

        if isinstance(data, PositionData):
            if data.direction == Direction.SHORT:
                direction: Direction = Direction.LONG
            elif data.direction == Direction.LONG:
                direction: Direction = Direction.SHORT
            else:       # Net position mode
                if data.volume > 0:
                    direction: Direction = Direction.SHORT
                else:
                    direction: Direction = Direction.LONG

            self.direction_combo.setCurrentIndex(
                self.direction_combo.findText(direction.value)
            )
            self.offset_combo.setCurrentIndex(
                self.offset_combo.findText(Offset.CLOSE.value)
            )
            self.volume_line.setText(str(abs(data.volume)))


class AboutDialog(MessageBoxBase):
    """
    Information about the trading platform.
    """

    def __init__(self, parent: QtWidgets.QWidget) -> None:
        """"""
        super().__init__(parent)

        self.init_ui()

    def init_ui(self) -> None:
        """"""
        self.title_label = SubtitleLabel("About VeighNa Evo", self)

        from vnpy import __version__ as vnpy_version
        from ... import __version__ as evo_version

        text: str = f"""
            By Traders, For Traders.

            Created by VeighNa Technology


            License：MIT
            Website：www.vnpy.com
            Github：www.github.com/vnpy/vnpy


            VeighNa Evo - {evo_version}
            VeighNa - {vnpy_version}
            Python - {platform.python_version()}
            PySide6 - {importlib_metadata.version("pyside6")}
            NumPy - {importlib_metadata.version("numpy")}
            pandas - {importlib_metadata.version("pandas")}
            """

        label: BodyLabel = BodyLabel()
        label.setText(text)
        label.setMinimumWidth(500)

        self.viewLayout.addWidget(self.title_label)
        self.viewLayout.addWidget(label)

        self.cancelButton.hide()


class GlobalDialog(MessageBoxBase):
    """
    Edit global setting.
    """

    def __init__(self, parent: QtWidgets.QWidget = None) -> None:
        """"""
        super().__init__(parent)

        self.widgets: dict[str, object] = {}

        self.init_ui()

    def init_ui(self) -> None:
        """"""
        self.title_label = SubtitleLabel("Global Configuration", self)

        settings: dict = copy(SETTINGS)
        settings.update(load_json(SETTING_FILENAME))

        # Initialize line edits and form layout based on setting.
        scroll_widget: QtWidgets.QWidget = QtWidgets.QWidget(parent=self)

        grid: QtWidgets.QFormLayout = QtWidgets.QGridLayout(parent=scroll_widget)
        row: int = 0

        for field_name, field_value in settings.items():
            if "datafeed" in field_name:
                continue

            field_type: type = type(field_value)
            widget: LineEdit = LineEdit()
            widget.setText(str(field_value))

            grid.addWidget(BodyLabel(f"{field_name} <{field_type.__name__}>"), row, 0)
            grid.addWidget(widget, row, 1)
            self.widgets[field_name] = (widget, field_type)

            row += 1

        scroll_widget.setLayout(grid)

        self.viewLayout.addWidget(self.title_label)
        self.viewLayout.addWidget(scroll_widget)

        self.yesButton.setText("Confirm")
        self.yesButton.clicked.connect(self.update_setting)

        self.widget.setFixedWidth(self.widget.width() * 6)

    def update_setting(self) -> None:
        """
        Get setting value from line edits and update global setting file.
        """
        settings: dict = {}
        for field_name, tp in self.widgets.items():
            widget, field_type = tp
            value_text: str = widget.text()

            if field_type == bool:
                if value_text == "True":
                    field_value: bool = True
                else:
                    field_value: bool = False
            else:
                field_value = field_type(value_text)

            settings[field_name] = field_value

        QtWidgets.QMessageBox.information(
            self,
            _("注意"),
            _("全局配置的修改需要重启后才会生效！"),
            QtWidgets.QMessageBox.Ok
        )

        save_json(SETTING_FILENAME, settings)
        self.accept()


class PivotWidgdet(QtWidgets.QWidget):
    """"""

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self.pivot = Pivot(self)
        self.stacked_widget = QtWidgets.QStackedWidget(self)
        self.vbox = QtWidgets.QVBoxLayout(self)

        self.vbox.addWidget(self.pivot, 0, QtCore.Qt.AlignLeft)
        self.vbox.addWidget(self.stacked_widget)
        self.vbox.setContentsMargins(0, 0, 0, 0)

        self.stacked_widget.currentChanged.connect(self.onCurrentIndexChanged)

    def add_widget(self, widget: QtWidgets.QWidget, name: str) -> None:
        """"""
        widget.setObjectName(name)

        self.stacked_widget.addWidget(widget)

        self.pivot.addItem(
            routeKey=name,
            text=name,
            onClick=lambda: self.stacked_widget.setCurrentWidget(widget)
        )

        if self.stacked_widget.count() == 1:
            self.stacked_widget.setCurrentWidget(widget)
            self.pivot.setCurrentItem(widget.objectName())

    def onCurrentIndexChanged(self, index: int) -> None:
        """"""
        widget: QtWidgets.QWidget = self.stacked_widget.widget(index)
        self.pivot.setCurrentItem(widget.objectName())



import ctypes
import platform
import sys
import traceback
import webbrowser
import types
import threading

from PySide6 import QtGui, QtWidgets, QtCore
from qfluentwidgets import (
    setTheme, Theme,
    PushButton, TextEdit, SubtitleLabel
)

from ..setting import SETTINGS
from ..utility import get_icon_path


Qt = QtCore.Qt
QtCore.pyqtSignal = QtCore.Signal
QtWidgets.QAction = QtGui.QAction
QtCore.QDate.toPyDate = QtCore.QDate.toPython
QtCore.QDateTime.toPyDate = QtCore.QDateTime.toPython


def create_qapp(app_name: str = "VeighNa Evo") -> QtWidgets.QApplication:
    """
    Create Qt Application.
    """
    qapp: QtWidgets.QApplication = QtWidgets.QApplication(sys.argv)
    setTheme(Theme.LIGHT)

    # Set up font
    font: QtGui.QFont = QtGui.QFont(SETTINGS["font.family"], SETTINGS["font.size"])
    qapp.setFont(font)

    # Set up icon
    icon: QtGui.QIcon = QtGui.QIcon(get_icon_path(__file__, "veighna.ico"))
    qapp.setWindowIcon(icon)

    # Set up windows process ID
    if "Windows" in platform.uname():
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(
            app_name
        )

    # Hide help button for all dialogs
    # qapp.setAttribute(QtCore.Qt.AA_DisableWindowContextHelpButton)

    # Exception Handling
    exception_widget: ExceptionWidget = ExceptionWidget()

    def excepthook(exctype: type, value: Exception, tb: types.TracebackType) -> None:
        """Show exception detail with QMessageBox."""
        sys.__excepthook__(exctype, value, tb)

        msg: str = "".join(traceback.format_exception(exctype, value, tb))
        exception_widget.signal.emit(msg)

    sys.excepthook = excepthook

    def threading_excepthook(args: threading.ExceptHookArgs) -> None:
        """Show exception detail from background threads with QMessageBox."""
        sys.__excepthook__(args.exc_type, args.exc_value, args.exc_traceback)

        msg: str = "".join(traceback.format_exception(args.exc_type, args.exc_value, args.exc_traceback))
        exception_widget.signal.emit(msg)

    threading.excepthook = threading_excepthook

    return qapp


class ExceptionWidget(QtWidgets.QWidget):
    """"""
    signal: QtCore.Signal = QtCore.Signal(str)

    def __init__(self, parent: QtWidgets.QWidget = None) -> None:
        """"""
        super().__init__(parent)

        self.init_ui()
        self.signal.connect(self.show_exception)

    def init_ui(self) -> None:
        """"""
        self.setFixedSize(600, 600)
        self.setWindowFlags(QtCore.Qt.FramelessWindowHint)
        self.setStyleSheet("background-color: rgb(243, 243, 243)")

        self.title_label = SubtitleLabel("Exception Triggered", self)

        self.msg_edit: TextEdit = TextEdit()
        self.msg_edit.setReadOnly(True)

        copy_button: PushButton = PushButton("Copy")
        copy_button.clicked.connect(self._copy_text)

        community_button: PushButton = PushButton("Help")
        community_button.clicked.connect(self._open_community)

        close_button: PushButton = PushButton("Close")
        close_button.clicked.connect(self.close)

        hbox: QtWidgets.QHBoxLayout = QtWidgets.QHBoxLayout()
        hbox.addWidget(copy_button)
        hbox.addWidget(community_button)
        hbox.addWidget(close_button)

        vbox: QtWidgets.QVBoxLayout = QtWidgets.QVBoxLayout()
        vbox.addWidget(self.title_label)
        vbox.addWidget(self.msg_edit)
        vbox.addLayout(hbox)

        self.setLayout(vbox)

    def show_exception(self, msg: str) -> None:
        """"""
        self.msg_edit.setText(msg)
        self.show()

    def _copy_text(self) -> None:
        """"""
        self.msg_edit.selectAll()
        self.msg_edit.copy()

    def _open_community(self) -> None:
        """"""
        webbrowser.open("https://www.vnpy.com/forum/forum/2-ti-wen-qiu-zhu")



from vnpy.trader.utility import *

from .constant import Exchange


def extract_vt_symbol(vt_symbol: str) -> tuple[str, Exchange]:
    """
    return (symbol, exchange)
    """
    symbol, exchange_str = vt_symbol.rsplit(".", 1)
    return symbol, Exchange(exchange_str)



from importlib import import_module
from types import ModuleType

from vnpy.trader.database import (
    DB_TZ,
    convert_tz,
    BarOverview,
    TickOverview,
    BaseDatabase
)

from .setting import SETTINGS


database: BaseDatabase = None


def get_database() -> BaseDatabase:
    """"""
    # Return database object if already inited
    global database
    if database:
        return database

    # Read database related global setting
    database_name: str = SETTINGS["database.name"]
    module_name: str = f"vnpy_{database_name}"

    # Try to import database module
    try:
        module: ModuleType = import_module(module_name)
    except ModuleNotFoundError:
        print(("Database adapter not found{}, use default DuckDB adpater").format(module_name))
        module: ModuleType = import_module("vnpy_duckdb")

    # Create database object from module
    database = module.Database()
    return database



import json
import ssl
import traceback
from threading import Thread
from typing import Optional

import websocket


class WebsocketClient:
    """
    Websocket API

    After creating the client object, use start() to run worker thread.
    The worker thread connects websocket automatically.

    Use stop to stop threads and disconnect websocket before destroying the client
    object (especially when exiting the programme).

    Default serialization format is json.

    Callbacks to overrides:
    * on_connected
    * on_disconnected
    * on_packet
    * on_error

    If you want to send anything other than JSON, override send_packet.
    """

    def __init__(self) -> None:
        """Constructor"""
        self.active: bool = False
        self.host: str = ""

        self.wsapp: websocket.WebSocketApp = None
        self.thread: Thread = None

        self.proxy_host: Optional[str] = None
        self.proxy_port: Optional[int] = None
        self.header: Optional[dict] = None
        self.ping_interval: int = 0
        self.receive_timeout: int = 0

        self.trace: bool = False

    def init(
        self,
        host: str,
        proxy_host: str = "",
        proxy_port: int = 0,
        ping_interval: int = 10,
        receive_timeout: int = 60,
        header: dict = None,
        trace: bool = False
    ) -> None:
        """
        :param host:
        :param proxy_host:
        :param proxy_port:
        :param header:
        :param ping_interval: unit: seconds, type: int
        """
        self.host = host
        self.ping_interval = ping_interval  # seconds
        self.receive_timeout = receive_timeout

        if header:
            self.header = header

        if proxy_host and proxy_port:
            self.proxy_host = proxy_host
            self.proxy_port = proxy_port

        websocket.enableTrace(trace)
        websocket.setdefaulttimeout(receive_timeout)

    def start(self) -> None:
        """
        Start the client and on_connected function is called after webscoket
        is connected succesfully.

        Please don't send packet untill on_connected fucntion is called.
        """
        self.active = True
        self.thread = Thread(target=self.run)
        self.thread.start()

    def stop(self) -> None:
        """
        Stop the client.
        """
        if not self.active:
            return

        self.active = False
        self.wsapp.close()

    def join(self) -> None:
        """
        Wait till all threads finish.

        This function cannot be called from worker thread or callback function.
        """
        self.thread.join()

    def send_packet(self, packet: dict) -> None:
        """
        Send a packet (dict data) to server

        override this if you want to send non-json packet
        """
        text: str = json.dumps(packet)
        self.wsapp.send(text)

    def run(self) -> None:
        """
        Keep running till stop is called.
        """
        def on_open(wsapp: websocket.WebSocketApp) -> None:
            self.on_connected()

        def on_close(wsapp: websocket.WebSocketApp, status_code: int, msg: str) -> None:
            self.on_disconnected(status_code, msg)

        def on_error(wsapp: websocket.WebSocketApp, e: Exception) -> None:
            self.on_error(e)

        def on_message(wsapp: websocket.WebSocketApp, message: str) -> None:
            self.on_message(message)

        self.wsapp = websocket.WebSocketApp(
            url=self.host,
            header=self.header,
            on_open=on_open,
            on_close=on_close,
            on_error=on_error,
            on_message=on_message
        )

        proxy_type: Optional[str] = None
        if self.proxy_host:
            proxy_type = "http"

        self.wsapp.run_forever(
            sslopt={"cert_reqs": ssl.CERT_NONE},
            ping_interval=self.ping_interval,
            http_proxy_host=self.proxy_host,
            http_proxy_port=self.proxy_port,
            proxy_type=proxy_type
        )

    def on_message(self, message: str) -> None:
        """
        Callback when weboscket app receives new message
        """
        self.on_packet(json.loads(message))

    def on_connected(self) -> None:
        """
        Callback when websocket is connected successfully.
        """
        pass

    def on_disconnected(self, status_code: int, msg: str) -> None:
        """
        Callback when websocket connection is closed.
        """
        pass

    def on_packet(packet: dict) -> None:
        """
        Callback when receiving data from server.
        """
        pass

    def on_error(self, e: Exception) -> None:
        """
        Callback when exception raised.
        """
        try:
            print("WebsocketClient on error" + "-" * 10)
            print(e)
        except Exception:
            traceback.print_exc()



from types import ModuleType
import webbrowser
from functools import partial
from importlib import import_module

from typing import Callable, Type

from qfluentwidgets import (
    FluentWindow, MessageBox,
    FluentIcon as FIF,
    PushButton, RoundMenu,
    Action, NavigationItemPosition,
)

from vnpy.event import EventEngine

import vnpy_evo
from .qt import QtCore, QtGui, QtWidgets
from .widget import (
    ConnectDialog,
    TradingWidget,
    AboutDialog,
    GlobalDialog,
    PivotWidgdet
)
from .monitor import (
    TickMonitor,
    OrderMonitor,
    TradeMonitor,
    PositionMonitor,
    AccountMonitor,
    LogMonitor,
    ActiveOrderMonitor,
    ContractManager
)
from ..app import BaseApp
from ..engine import MainEngine
from ..utility import get_icon_path, TRADER_DIR


class MainWindow(FluentWindow):
    """
    Main window of the trading platform.
    """

    def __init__(self, main_engine: MainEngine, event_engine: EventEngine) -> None:
        """"""
        super().__init__()

        self.main_engine: MainEngine = main_engine
        self.event_engine: EventEngine = event_engine

        self.window_title: str = f"VeighNa Evo - {vnpy_evo.__version__}  [{TRADER_DIR}]"

        self.app_widgets: dict[str, QtWidgets.QWidget] = {}

        self.init_ui()

    def init_ui(self) -> None:
        """"""
        self.setWindowTitle(self.window_title)

        icon: QtGui.QIcon = QtGui.QIcon(get_icon_path(__file__, "veighna.ico"))
        self.setWindowIcon(icon)

        self.init_widgets()
        self.init_navigation()

    def init_widgets(self) -> None:
        """"""
        self.home_widget = HomeWidget(self.main_engine, self.event_engine)
        self.home_widget.setObjectName("home")

        self.contract_manager = ContractManager(self.main_engine, self.event_engine)
        self.contract_manager.setObjectName("contract")

        all_apps: list[BaseApp] = self.main_engine.get_all_apps()
        for app in all_apps:
            ui_module: ModuleType = import_module(app.app_module + ".ui")
            widget_class: Type = getattr(ui_module, app.widget_name)
            widget: QtWidgets.QWidget = widget_class(self.main_engine, self.event_engine)
            widget.setObjectName(app.display_name)
            self.app_widgets[app.display_name] = widget

    def init_navigation(self) -> None:
        """"""
        self.addSubInterface(self.home_widget, FIF.HOME, "Home")
        self.addSubInterface(self.contract_manager, FIF.SEARCH, "Find contract")

        for name, widget in self.app_widgets.items():
            self.addSubInterface(widget, FIF.ROBOT, name)

        self.navigationInterface.addItem(
            routeKey="froum",
            icon=FIF.HELP,
            text="Community forum",
            onClick=self.open_forum,
            selectable=False,
            position=NavigationItemPosition.BOTTOM
        )

        self.navigationInterface.addItem(
            routeKey="github",
            icon=FIF.GITHUB,
            text="Github",
            onClick=self.open_github,
            selectable=False,
            position=NavigationItemPosition.BOTTOM
        )

        self.navigationInterface.addItem(
            routeKey="setting",
            icon=FIF.SETTING,
            text="Settings",
            onClick=self.edit_global_setting,
            selectable=False,
            position=NavigationItemPosition.BOTTOM
        )

        self.navigationInterface.addItem(
            routeKey="about",
            icon=FIF.INFO,
            text="About",
            onClick=self.open_about,
            selectable=False,
            position=NavigationItemPosition.BOTTOM
        )

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        """
        Call main engine close function before exit.
        """
        msgbox: MessageBox = MessageBox("Notice", "Do you confirm exit?", self.window())
        reply: int = msgbox.exec()

        if reply:
            self.main_engine.close()

            event.accept()
        else:
            event.ignore()

    def open_forum(self) -> None:
        """
        """
        webbrowser.open("https://www.vnpy.com/forum/")

    def open_github(self) -> None:
        """
        """
        webbrowser.open("https://github.com/veighna-global/vnpy_evo")

    def edit_global_setting(self) -> None:
        """
        """
        dialog: GlobalDialog = GlobalDialog(self)
        dialog.exec()

    def open_about(self) -> None:
        """
        Open contract manager.
        """
        dialog: AboutDialog = AboutDialog(self)
        dialog.exec()


class HomeWidget(QtWidgets.QWidget):
    """"""

    def __init__(self, main_engine: MainEngine, event_engine: EventEngine) -> None:
        """"""
        super().__init__()

        self.main_engine: MainEngine = main_engine
        self.event_engine: EventEngine = event_engine

        self.init_ui()
        self.init_menu()

    def init_ui(self) -> None:
        """"""
        # Create widgets
        self.trading_widget = TradingWidget(self.main_engine, self.event_engine)
        self.tick_monitor = TickMonitor(self.main_engine, self.event_engine)
        self.order_monitor = OrderMonitor(self.main_engine, self.event_engine)
        self.active_monitor = ActiveOrderMonitor(self.main_engine, self.event_engine)
        self.trade_monitor = TradeMonitor(self.main_engine, self.event_engine)
        self.position_monitor = PositionMonitor(self.main_engine, self.event_engine)
        self.account_monitor = AccountMonitor(self.main_engine, self.event_engine)
        self.log_monitor = LogMonitor(self.main_engine, self.event_engine)

        self.menu: RoundMenu = RoundMenu(parent=self)

        self.menu_button: PushButton = PushButton("Connect Gateway")
        self.menu_button.clicked.connect(self.show_menu)

        # Set layout
        mid_pivot = PivotWidgdet(self)
        mid_pivot.add_widget(self.active_monitor, "Open Orders")
        mid_pivot.add_widget(self.order_monitor, "Order History")

        bottom_pivot = PivotWidgdet(self)
        bottom_pivot.add_widget(self.log_monitor, "Log")
        bottom_pivot.add_widget(self.trade_monitor, "Trade History")
        bottom_pivot.add_widget(self.position_monitor, "Positions")
        bottom_pivot.add_widget(self.account_monitor, "Assets")

        vbox1 = QtWidgets.QVBoxLayout()
        vbox1.addWidget(self.tick_monitor)
        vbox1.addWidget(mid_pivot)
        vbox1.addWidget(bottom_pivot)

        vbox2 = QtWidgets.QVBoxLayout()
        vbox2.addWidget(self.trading_widget)
        vbox2.addWidget(self.menu_button)

        hbox = QtWidgets.QHBoxLayout()
        hbox.addLayout(vbox2)
        hbox.addLayout(vbox1)

        self.setLayout(hbox)

        # Connect signal
        self.tick_monitor.itemDoubleClicked.connect(self.trading_widget.update_with_cell)
        self.position_monitor.itemDoubleClicked.connect(self.trading_widget.update_with_cell)

    def init_menu(self) -> None:
        """"""
        gateway_names: list = self.main_engine.get_all_gateway_names()
        for name in gateway_names:
            func: Callable = partial(self.connect_gateway, name)

            action = Action(f"Connect {name}")
            action.triggered.connect(func)

            self.menu.addAction(action)

    def show_menu(self) -> None:
        """"""
        pos = self.menu_button.mapToGlobal(QtCore.QPoint(self.menu_button.width() + 5, 0))
        self.menu.exec(pos, ani=True)

    def connect_gateway(self, gateway_name: str) -> None:
        """
        Open connect dialog for gateway connection.
        """
        dialog: ConnectDialog = ConnectDialog(self.main_engine, gateway_name, self)
        dialog.exec()



from enum import Enum

from vnpy.trader.constant import (
    Direction,
    Offset,
    Status,
    Product,
    OrderType,
    OptionType,
    Currency,
    Interval
)


class Exchange(Enum):
    """
    Exchange.
    """
    # Crypto
    BINANCE = "BINANCE"
    OKX = "OKX"
    BYBIT = "BYBIT"
    BTSE = "BTSE"
    DERIBIT = "DERIBIT"

    # Global
    OTC = "OTC"

    # Special Function
    LOCAL = "LOCAL"         # For local generated data



from datetime import datetime
from threading import Thread
from queue import Queue, Empty

import requests
from loguru import logger

from vnpy.trader.engine import (
    BaseEngine,
    EventEngine,
    MainEngine as OriginalMainEngine,
    OmsEngine,
    EmailEngine,
    Event,
    EVENT_LOG,
    Path,
    get_folder_path
)

from .object import LogData
from .setting import SETTINGS


class MainEngine(OriginalMainEngine):
    """New main engine of VeighNa Evo"""

    def init_engines(self) -> None:
        """
        Init all engines.
        """
        self.add_engine(LogEngine)
        self.add_engine(OmsEngine)
        self.add_engine(EmailEngine)
        self.add_engine(TelegramEngine)


class LogEngine(BaseEngine):
    """Use loguru instead of logging"""

    def __init__(self, main_engine: MainEngine, event_engine: EventEngine) -> None:
        """"""
        super().__init__(main_engine, event_engine, "log")

        self.active = SETTINGS["log.active"]
        self.level: int = SETTINGS["log.level"]
        self.format: str = "{time}  {level}: {message}"

        if not SETTINGS["log.console"]:
            logger.remove()     # Remove default stderr output

        if SETTINGS["log.file"]:
            today_date: str = datetime.now().strftime("%Y%m%d")
            filename: str = f"vt_{today_date}.log"
            log_path: Path = get_folder_path("log")
            file_path: Path = log_path.joinpath(filename)

            logger.add(
                sink=file_path,
                level=self.level,
                retention="4 weeks"
            )

        self.register_event()

    def register_event(self) -> None:
        """Register log event handler"""
        self.event_engine.register(EVENT_LOG, self.process_log_event)

    def process_log_event(self, event: Event) -> None:
        """Process log event"""
        if not self.active:
            return

        log: LogData = event.data
        logger.log(log.level, log.msg)


class TelegramEngine(BaseEngine):
    """Telegram message sending engine"""

    def __init__(self, main_engine: MainEngine, event_engine: EventEngine) -> None:
        super().__init__(main_engine, event_engine, "telegram")

        self.active: bool = SETTINGS.get("telegram.active", False)
        self.token: str = SETTINGS.get("telegram.token", "")
        self.chat: str = SETTINGS.get("telegram.chat", "")
        self.url: str = f"https://api.telegram.org/bot{self.token}/sendMessage"

        self.proxies: dict[str, str] = {}
        proxy: str = SETTINGS.get("telegram.proxy", "")
        if proxy:
            self.proxies["http"] = proxy
            self.proxies["https"] = proxy

        self.thread: Thread = Thread(target=self.run, daemon=True)
        self.queue: Queue = Queue()

        if self.active:
            self.register_event()
            self.thread.start()

    def register_event(self) -> None:
        """Register event handler"""
        self.event_engine.register(EVENT_LOG, self.process_log_event)

    def process_log_event(self, event: Event) -> None:
        """Process log event"""
        log: LogData = event.data

        msg = f"{log.time}\t[{log.gateway_name}] {log.msg}"
        self.queue.put(msg)

    def close(self) -> None:
        """Stop task thread"""
        if not self.active:
            return

        self.active = False
        self.thread.join()

    def run(self) -> None:
        """Task thread loop"""
        while self.active:
            try:
                msg: str = self.queue.get(block=True, timeout=1)
                self.send_msg(msg)
            except Empty:
                pass

    def send_msg(self, msg: str) -> dict:
        """Sending message"""
        data: dict = {
            "chat_id": self.chat,
            "text": msg
        }

        # 发送请求
        try:
            r: requests.Response = requests.post(self.url, data=data)
            return r.json()
        except Exception:
            return None
